
//
//  MComp.cpp - Packed Motion Compensation
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 21 Feb, 2002 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules
    
#include "MComp.h"
#include "MCompEx.h"



vmp_MCompProc	vmp_MCompHz8R;
vmp_MCompProc	vmp_MCompHz16R;
vmp_MCompProc	vmp_MCompVt8R;
vmp_MCompProc	vmp_MCompVt16R;





//
//	error Bit Masking Values
//
#if 01
	// Values in memory are more efficient then 32bit definitions.
		
unsigned int eMASK01 = 0x01010101;		// Error correction Mask 01
unsigned int eMASK02 = 0x02020202;		// Error correction Mask 02
unsigned int eMASK03 = 0x03030303;		// Error correction Mask 03
unsigned int eMASK3F = 0x3f3f3f3f;		// Error correction Mask 3f
unsigned int eMASK7F = 0x7f7f7f7f;		// Error correction Mask 7f
#else
#define eMASK01	0x01010101				// Error correction Mask 01
#define eMASK02 0x02020202				// Error correction Mask 02
#define eMASK03	0x03030303				// Error correction Mask 03
#define eMASK3F	0x3f3f3f3f				// Error correction Mask 3f
#define eMASK7F	0x7f7f7f7f				// Error correction Mask 7f
#endif



//
//	 8x8 Horizontal Motion Compensation w/Rounding  ***SLOW***
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][X]
//	[ ][ ]

void vmp_MCompHz8RGenericSLOW( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{
	uint dy, dx;

	ASSERT_PTR(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

	for (dy = 0; dy < 8; dy++)
	  {
		for (dx = 0; dx < 8; dx++)
		  {
			pbDst[dx] = (unsigned char) ((pbSrc[dx] + pbSrc[dx+1] + 1) >> 1);
		  }

		pbSrc += iStride;		// +- to beginning next scanline!
		pbDst += iStride;
	}
}



//
//	 8x8 Horizontal Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][X]
//	[ ][ ]

void vmp_MCompHz8RGeneric( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{
	register uint32 c0, c1, d0, d1;
	const uint32 *pSrc;
	uint32 *pDst;
	uint dy;

	ASSERT_PTR(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

	iStride >>= 2;		// # of 32bit words

	pSrc = (uint32*) pbSrc;
	pDst = (uint32*) pbDst;

		//	Endian-Less Quad 4x8bit Averager (a+b+1)/2  (+1) use '|' OR error correction
		//	No Carry between 8bit sections!  Handles two 32bit sections.
		//	Uses Law of Distribution:  a(b+c)=ab+ac

	dy = 8;
	do {	// Horizontal Interpolation with rounding
		d0 = *pSrc;
		d1 = *(pSrc+1);

#ifdef VMP_BIG_ENDIAN					// Big Endian
		c0 = (d0 << 8) | (d1>>24);
		c1 = (d1 << 8) | (*(pSrc+2)>>24);
#else									// Little Endian
		c0 = (d0 >> 8) | (d1<<24);
		c1 = (d1 >> 8) | (*(pSrc+2)<<24);
#endif
	
		*pDst =     (( c0 >> 1 ) & eMASK7F) + ((d0 >> 1) & eMASK7F) + ((d0 | c0) & eMASK01);
		*(pDst+1) = (( c1 >> 1 ) & eMASK7F) + ((d1 >> 1) & eMASK7F) + ((d1 | c1) & eMASK01);

		pSrc += iStride;
		pDst += iStride;
	} while (--dy);
}



//
//	 16x16 Horizontal Motion Compensation w/Rounding  ***SLOW***
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][X]
//	[ ][ ]

void vmp_MCompHz16RGenericSLOW( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{
	uint dy, dx;

	ASSERT_PTR(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

	for (dy = 0; dy < 16; dy++)
	  {
		for (dx = 0; dx < 16; dx++)
		  {
			pbDst[dx] = (unsigned char) ((pbSrc[dx] + pbSrc[dx+1] + 1) >> 1);
		  }

		pbSrc += iStride;		// +- to beginning next scanline!
		pbDst += iStride;
	}
}



//
//	 16x16 Horizontal Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][X]
//	[ ][ ]

void vmp_MCompHz16RGeneric( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{
	register uint32 c0, c1, c2, c3, d0, d1, d2, d3;
	const uint32 *pSrc;
	uint32 *pDst;
	uint dy;

	ASSERT_PTR(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

	iStride >>= 2;		// # of 32bit words

	pSrc = (uint32*) pbSrc;
	pDst = (uint32*) pbDst;

		//	Endian-Less Quad 4x8bit Averager (a+b+1)/2  (+1) use '|' OR error correction
		//	No Carry between 8bit sections!  Handles two 32bit sections.
		//	Uses Law of Distribution:  a(b+c)=ab+ac

	dy = 16;
	do {	// Horizontal Interpolation with rounding
		d0 = *pSrc;
		d1 = *(pSrc+1);
		d2 = *(pSrc+2);
		d3 = *(pSrc+3);

#ifdef VMP_BIG_ENDIAN					// Big Endian
		c0 = (d0 << 8) | (d1>>24);
		c1 = (d1 << 8) | (d2>>24);
		c2 = (d2 << 8) | (d3>>24);
		c3 = (d3 << 8) | (*(pSrc+4)>>24);
#else									// Little Endian
		c0 = (d0 >> 8) | (d1<<24);
		c1 = (d1 >> 8) | (d2<<24);
		c2 = (d2 >> 8) | (d3<<24);
		c3 = (d3 >> 8) | (*(pSrc+4)<<24);
#endif
	
		*pDst =     (( c0 >> 1 ) & eMASK7F) + ((d0 >> 1) & eMASK7F) + ((d0 | c0) & eMASK01);
		*(pDst+1) = (( c1 >> 1 ) & eMASK7F) + ((d1 >> 1) & eMASK7F) + ((d1 | c1) & eMASK01);
		*(pDst+2) = (( c2 >> 1 ) & eMASK7F) + ((d2 >> 1) & eMASK7F) + ((d2 | c2) & eMASK01);
		*(pDst+3) = (( c3 >> 1 ) & eMASK7F) + ((d3 >> 1) & eMASK7F) + ((d3 | c3) & eMASK01);

		pSrc += iStride;
		pDst += iStride;
	} while (--dy);
}



//_______________________________________________________________________________________



//
//	 8x8 Vertical Motion Compensation w/Rounding  ***SLOW***
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][ ]
//	[X][ ]

void vmp_MCompVt8RGenericSLOW( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{
	uint dy;
	int dx;

	ASSERT_PTR(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

	for (dy = 0; dy < 8; dy++)
	  {
		for (dx = 0; dx < 8; dx++)
		  {
			pbDst[dx] = (unsigned char) ((pbSrc[dx] + pbSrc[dx+iStride] + 1) >> 1);
		  }

		pbSrc += iStride;		// +- to beginning next scanline!
		pbDst += iStride;
	}
}



//
//	 8x8 Vertical Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][ ]
//	[X][ ]

void vmp_MCompVt8RGeneric( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{
	const uint32 *pS;
	const uint32 *pR;
	uint32 *pDst;
	uint dy;

	ASSERT_PTR(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

	iStride >>= 2;		// # of 32bit words

	pS = (uint32*) pbSrc;
	pR = pS + iStride;			
	pDst = (uint32*) pbDst;

		//	Endian-Less Quad 4x8bit Averager (a+b+1)/2  (+1) use '|' OR error correction
		//	No Carry between 8bit sections!  Handles two 32bit sections.
		//	Uses Law of Distribution:  a(b+c)=ab+ac

	dy = 8;
	do {	// Vertical Interpolation with rounding
		*(pDst+0) = ((*pS     >> 1)      & eMASK7F) 
		          + ((*pR     >> 1)      & eMASK7F) 
		          + ((*pS     | *pR)     & eMASK01);
		*(pDst+1) = ((*(pS+1) >> 1)      & eMASK7F) 
		          + ((*(pR+1) >> 1)      & eMASK7F) 
		          + ((*(pS+1) | *(pR+1)) & eMASK01);
		pS = pR;
		pR   += iStride;
		pDst += iStride;
	} while (--dy);
}



//
//	 16x16 Vertical Motion Compensation w/Rounding  ***SLOW***
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][ ]
//	[X][ ]

void vmp_MCompVt16RGenericSLOW( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{
	uint dy;
	int dx;

	ASSERT_PTR(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

	for (dy = 0; dy < 16; dy++)
	  {
		for (dx = 0; dx < 16; dx++)
		  {
			pbDst[dx] = (unsigned char) ((pbSrc[dx] + pbSrc[dx+iStride] + 1) >> 1);
		  }

		pbSrc += iStride;		// +- to beginning next scanline!
		pbDst += iStride;
	}
}



//
//	 16x16 Vertical Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][ ]
//	[X][ ]

void vmp_MCompVt16RGeneric( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{
	const uint32 *pS;
	const uint32 *pR;
	uint32 *pDst;
	uint dy;

	ASSERT_PTR(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

	iStride >>= 2;		// # of 32bit words

	pS = (uint32*) pbSrc;
	pR = pS + iStride;			
	pDst = (uint32*) pbDst;

		//	Endian-Less Quad 4x8bit Averager (a+b+1)/2  (+1) use '|' OR error correction
		//	No Carry between 8bit sections!  Handles two 32bit sections.
		//	Uses Law of Distribution:  a(b+c)=ab+ac

	dy = 16;
	do {	// Vertical Interpolation with rounding
		*(pDst+0) = ((*pS     >> 1)      & eMASK7F) 
		          + ((*pR     >> 1)      & eMASK7F) 
		          + ((*pS     | *pR)     & eMASK01);
		*(pDst+1) = ((*(pS+1) >> 1)      & eMASK7F) 
		          + ((*(pR+1) >> 1)      & eMASK7F) 
		          + ((*(pS+1) | *(pR+1)) & eMASK01);
		*(pDst+2) = ((*(pS+2) >> 1)      & eMASK7F) 
		          + ((*(pR+2) >> 1)      & eMASK7F) 
		          + ((*(pS+2) | *(pR+2)) & eMASK01);
		*(pDst+3) = ((*(pS+3) >> 1)      & eMASK7F) 
		          + ((*(pR+3) >> 1)      & eMASK7F) 
		          + ((*(pS+3) | *(pR+3)) & eMASK01);
		pS = pR;
		pR   += iStride;
		pDst += iStride;
	} while (--dy);
}




